@@ -1,23 +1,23 @@
+from __future__ import annotations
+
 import copy
 import functools
 import logging
 import math
+import operator
 import re
 from contextlib import nullcontext as does_not_raise
 
 import pytest
 
-from pint import (
-    DefinitionSyntaxError,
-    DimensionalityError,
-    RedefinitionError,
-    UndefinedUnitError,
-)
+from pint import DimensionalityError, RedefinitionError, UndefinedUnitError, errors
 from pint.compat import np
 from pint.registry import LazyRegistry, UnitRegistry
 from pint.testsuite import QuantityTestCase, assert_no_warnings, helpers
 from pint.util import ParserHelper, UnitsContainer
 
+from .helpers import internal
+
 
 # TODO: do not subclass from QuantityTestCase
 class TestUnit(QuantityTestCase):
@@ -59,6 +59,33 @@ class TestUnit(QuantityTestCase):
             with subtests.test(spec):
                 assert spec.format(x) == result
 
+    def test_latex_escaping(self, subtests):
+        ureg = UnitRegistry()
+        ureg.define(r"percent = 1e-2 = %")
+        x = ureg.Unit(UnitsContainer(percent=1))
+        for spec, result in {
+            "L": r"\mathrm{percent}",
+            "L~": r"\mathrm{\%}",
+            "Lx": r"\si[]{\percent}",
+            "Lx~": r"\si[]{\%}",
+        }.items():
+            with subtests.test(spec):
+                ureg.default_format = spec
+                assert f"{x}" == result, f"Failed for {spec}, got {x} expected {result}"
+        # no '#' here as it's a comment char when define()ing new units
+        ureg.define(r"weirdunit = 1 = \~_^&%$_{}")
+        x = ureg.Unit(UnitsContainer(weirdunit=1))
+        for spec, result in {
+            "L": r"\mathrm{weirdunit}",
+            "L~": r"\mathrm{\textbackslash \textasciitilde \_\textasciicircum \&\%\$\_\{\}}",
+            "Lx": r"\si[]{\weirdunit}",
+            # TODO: Currently makes \si[]{\\~_^&%$_{}} (broken). What do we even want this to be?
+            # "Lx~": r"\si[]{\textbackslash \textasciitilde \_\textasciicircum \&\%\$\_\{\}}",
+        }.items():
+            with subtests.test(spec):
+                ureg.default_format = spec
+                assert f"{x}" == result, f"Failed for {spec}, {result}"
+
     def test_unit_default_formatting(self, subtests):
         ureg = UnitRegistry()
         x = ureg.Unit(UnitsContainer(meter=2, kilogram=1, second=-1))
@@ -80,6 +107,7 @@ class TestUnit(QuantityTestCase):
                 ureg.default_format = spec
                 assert f"{x}" == result, f"Failed for {spec}, {result}"
 
+    @pytest.mark.xfail(reason="Still not clear how default formatting will work.")
     def test_unit_formatting_defaults_warning(self):
         ureg = UnitRegistry()
         ureg.default_format = "~P"
@@ -112,17 +140,18 @@ class TestUnit(QuantityTestCase):
                 assert f"{x}" == result, f"Failed for {spec}, {result}"
 
     def test_unit_formatting_custom(self, monkeypatch):
-        from pint import formatting, register_unit_format
-
-        monkeypatch.setattr(formatting, "_FORMATTERS", formatting._FORMATTERS.copy())
+        from pint import register_unit_format
+        from pint.delegates.formatter._spec_helpers import REGISTERED_FORMATTERS
 
         @register_unit_format("new")
-        def format_new(unit, **options):
+        def format_new(unit, *args, **options):
             return "new format"
 
         ureg = UnitRegistry()
 
-        assert "{:new}".format(ureg.m) == "new format"
+        assert f"{ureg.m:new}" == "new format"
+
+        del REGISTERED_FORMATTERS["new"]
 
     def test_ipython(self):
         alltext = []
@@ -132,6 +161,13 @@ class TestUnit(QuantityTestCase):
             def text(text):
                 alltext.append(text)
 
+            @classmethod
+            def pretty(cls, data):
+                try:
+                    data._repr_pretty_(cls, False)
+                except AttributeError:
+                    alltext.append(str(data))
+
         ureg = UnitRegistry()
         x = ureg.Unit(UnitsContainer(meter=2, kilogram=1, second=-1))
         assert x._repr_html_() == "kilogram meter<sup>2</sup>/second"
@@ -171,7 +207,7 @@ class TestUnit(QuantityTestCase):
         ("unit", "power_ratio", "expectation", "expected_unit"),
         [
             ("m", 2, does_not_raise(), "m**2"),
-            ("m", dict(), pytest.raises(TypeError), None),
+            ("m", {}, pytest.raises(TypeError), None),
         ],
     )
     def test_unit_pow(self, unit, power_ratio, expectation, expected_unit):
@@ -217,7 +253,6 @@ class TestUnit(QuantityTestCase):
         assert not (self.U_("byte") != self.U_("byte"))
 
     def test_unit_cmp(self):
-
         x = self.U_("m")
         assert x < self.U_("km")
         assert x > self.U_("mm")
@@ -227,17 +262,14 @@ class TestUnit(QuantityTestCase):
         assert y < 1e6
 
     def test_dimensionality(self):
-
         x = self.U_("m")
         assert x.dimensionality == UnitsContainer({"[length]": 1})
 
     def test_dimensionless(self):
-
         assert self.U_("m/mm").dimensionless
         assert not self.U_("m").dimensionless
 
     def test_unit_casting(self):
-
         assert int(self.U_("m/mm")) == 1000
         assert float(self.U_("mm/m")) == 1e-3
         assert complex(self.U_("mm/mm")) == 1 + 0j
@@ -260,12 +292,12 @@ class TestRegistry(QuantityTestCase):
         cls.ureg.autoconvert_offset_to_baseunit = False
 
     def test_base(self):
-        ureg = UnitRegistry(None)
+        ureg = UnitRegistry(None, on_redefinition="raise")
         ureg.define("meter = [length]")
-        with pytest.raises(DefinitionSyntaxError):
+        with pytest.raises(errors.RedefinitionError):
             ureg.define("meter = [length]")
         with pytest.raises(TypeError):
-            ureg.define(list())
+            ureg.define([])
         ureg.define("degC = kelvin; offset: 273.15")
 
     def test_define(self):
@@ -274,15 +306,15 @@ class TestRegistry(QuantityTestCase):
         assert len(dir(ureg)) > 0
 
     def test_load(self):
-        import pkg_resources
+        from importlib.resources import files
 
-        from pint import unit
+        from .. import compat
 
-        data = pkg_resources.resource_filename(unit.__name__, "default_en.txt")
+        data = files(compat.__package__).joinpath("default_en.txt")
         ureg1 = UnitRegistry()
         ureg2 = UnitRegistry(data)
         assert dir(ureg1) == dir(ureg2)
-        with pytest.raises(ValueError):
+        with pytest.raises(FileNotFoundError):
             UnitRegistry(None).load_definitions("notexisting")
 
     def test_default_format(self):
@@ -375,6 +407,16 @@ class TestRegistry(QuantityTestCase):
             1, UnitsContainer(meter=37, second=-4.321)
         )
 
+    def test_parse_pretty_degrees(self):
+        for exp in ("1Δ°C", "1 Δ°C", "ΔdegC", "delta_°C"):
+            assert self.ureg.parse_expression(exp) == self.Q_(
+                1, UnitsContainer(delta_degree_Celsius=1)
+            )
+        assert self.ureg.parse_expression("")
+        assert self.ureg.parse_expression("mol °K") == self.Q_(
+            1, UnitsContainer(mol=1, kelvin=1)
+        )
+
     def test_parse_factor(self):
         assert self.ureg.parse_expression("42*meter") == self.Q_(
             42, UnitsContainer(meter=1.0)
@@ -538,8 +580,7 @@ class TestRegistry(QuantityTestCase):
         assert f3(3.0 * ureg.centimeter) == 0.03 * ureg.centimeter
         assert f3(3.0 * ureg.meter) == 3.0 * ureg.centimeter
 
-        def gfunc(x, y):
-            return x + y
+        gfunc = operator.add
 
         g0 = ureg.wraps(None, [None, None])(gfunc)
         assert g0(3, 1) == 4
@@ -565,12 +606,27 @@ class TestRegistry(QuantityTestCase):
         h3 = ureg.wraps((None,), (None, None))(hfunc)
         assert h3(3, 1) == (3, 1)
 
-    def test_wrap_referencing(self):
+        def kfunc(a, /, b, c=5, *, d=6):
+            return a, b, c, d
 
+        k1 = ureg.wraps((None,), (None, None, None, None))(kfunc)
+        assert k1(1, 2, 3, d=4) == (1, 2, 3, 4)
+        assert k1(1, 2, c=3, d=4) == (1, 2, 3, 4)
+        assert k1(1, b=2, c=3, d=4) == (1, 2, 3, 4)
+        assert k1(1, d=4, b=2, c=3) == (1, 2, 3, 4)
+        assert k1(1, 2, c=3) == (1, 2, 3, 6)
+        assert k1(1, 2, d=4) == (1, 2, 5, 4)
+        assert k1(1, 2) == (1, 2, 5, 6)
+
+        k2 = ureg.wraps((None,), ("meter", "centimeter", "meter", "centimeter"))(kfunc)
+        assert k2(
+            1 * ureg.meter, 2 * ureg.centimeter, 3 * ureg.meter, d=4 * ureg.centimeter
+        ) == (1, 2, 3, 4)
+
+    def test_wrap_referencing(self):
         ureg = self.ureg
 
-        def gfunc(x, y):
-            return x + y
+        gfunc = operator.add
 
         def gfunc2(x, y):
             return x**2 + y
@@ -615,6 +671,7 @@ class TestRegistry(QuantityTestCase):
         assert f0(3.0 * ureg.centimeter) == 0.03 * ureg.meter
         with pytest.raises(DimensionalityError):
             f0(3.0 * ureg.kilogram)
+        assert f0(x=3.0 * ureg.centimeter) == 0.03 * ureg.meter
 
         f0b = ureg.check(ureg.meter)(func)
         with pytest.raises(DimensionalityError):
@@ -623,14 +680,13 @@ class TestRegistry(QuantityTestCase):
         with pytest.raises(DimensionalityError):
             f0b(3.0 * ureg.kilogram)
 
-        def gfunc(x, y):
-            return x / y
+        gfunc = operator.truediv
 
         g0 = ureg.check(None, None)(gfunc)
         assert g0(6, 2) == 3
         assert g0(6 * ureg.parsec, 2) == 3 * ureg.parsec
 
-        g1 = ureg.check("[speed]", "[time]")(gfunc)
+        g1 = ureg.check("[velocity]", "[time]")(gfunc)
         with pytest.raises(DimensionalityError):
             g1(3.0, 1)
         with pytest.raises(DimensionalityError):
@@ -643,22 +699,22 @@ class TestRegistry(QuantityTestCase):
         )
 
         with pytest.raises(TypeError):
-            ureg.check("[speed]")(gfunc)
+            ureg.check("[velocity]")(gfunc)
         with pytest.raises(TypeError):
-            ureg.check("[speed]", "[time]", "[mass]")(gfunc)
+            ureg.check("[velocity]", "[time]", "[mass]")(gfunc)
 
     def test_to_ref_vs_to(self):
         self.ureg.autoconvert_offset_to_baseunit = True
         q = 8.0 * self.ureg.inch
         t = 8.0 * self.ureg.degF
         dt = 8.0 * self.ureg.delta_degF
-        assert q.to("yard").magnitude == self.ureg._units[
+        assert q.to("yard").magnitude == internal(self.ureg)._units[
             "inch"
         ].converter.to_reference(8.0)
-        assert t.to("kelvin").magnitude == self.ureg._units[
+        assert t.to("kelvin").magnitude == internal(self.ureg)._units[
             "degF"
         ].converter.to_reference(8.0)
-        assert dt.to("kelvin").magnitude == self.ureg._units[
+        assert dt.to("kelvin").magnitude == internal(self.ureg)._units[
             "delta_degF"
         ].converter.to_reference(8.0)
 
@@ -668,7 +724,7 @@ class TestRegistry(QuantityTestCase):
         with caplog.at_level(logging.DEBUG):
             d("meter = [fruits]")
             d("kilo- = 1000")
-            d("[speed] = [vegetables]")
+            d("[velocity] = [vegetables]")
 
             # aliases
             d("bla = 3.2 meter = inch")
@@ -777,7 +833,6 @@ class TestRegistry(QuantityTestCase):
 
 
 class TestCaseInsensitiveRegistry(QuantityTestCase):
-
     kwargs = dict(case_sensitive=False)
 
     def test_case_sensitivity(self):
@@ -819,7 +874,6 @@ class TestCompatibleUnits(QuantityTestCase):
         self._test(self.ureg.kelvin)
 
     def test_context_sp(self):
-
         gd = self.ureg.get_dimensionality
 
         # length, frequency, energy
@@ -858,13 +912,6 @@ class TestCompatibleUnits(QuantityTestCase):
 
 
 class TestRegistryWithDefaultRegistry(TestRegistry):
-    @classmethod
-    def setup_class(cls):
-        from pint import _DEFAULT_REGISTRY
-
-        cls.ureg = _DEFAULT_REGISTRY
-        cls.Q_ = cls.ureg.Quantity
-
     def test_lazy(self):
         x = LazyRegistry()
         x.test = "test"
@@ -873,19 +920,21 @@ class TestRegistryWithDefaultRegistry(TestRegistry):
         y("meter")
         assert isinstance(y, UnitRegistry)
 
-    def test_redefinition(self):
-        d = self.ureg.define
-        with pytest.raises(DefinitionSyntaxError):
+    def test_redefinition(self, func_registry):
+        ureg = UnitRegistry(on_redefinition="raise")
+        d = ureg.define
+        assert "meter" in internal(self.ureg)._units
+        with pytest.raises(RedefinitionError):
             d("meter = [time]")
         with pytest.raises(RedefinitionError):
             d("meter = [newdim]")
         with pytest.raises(RedefinitionError):
             d("kilo- = 1000")
         with pytest.raises(RedefinitionError):
-            d("[speed] = [length]")
+            d("[velocity] = [length]")
 
         # aliases
-        assert "inch" in self.ureg._units
+        assert "inch" in internal(self.ureg)._units
         with pytest.raises(RedefinitionError):
             d("bla = 3.2 meter = inch")
         with pytest.raises(RedefinitionError):
@@ -894,7 +943,6 @@ class TestRegistryWithDefaultRegistry(TestRegistry):
 
 # TODO: remove QuantityTestCase
 class TestConvertWithOffset(QuantityTestCase):
-
     # The dicts in convert_with_offset are used to create a UnitsContainer.
     # We create UnitsContainer to avoid any auto-conversion of units.
     convert_with_offset = [
@@ -941,6 +989,8 @@ class TestConvertWithOffset(QuantityTestCase):
         (({"degC": 2}, {"kelvin": 2}), "error"),
         (({"degC": 1, "degF": 1}, {"kelvin": 2}), "error"),
         (({"degC": 1, "kelvin": 1}, {"kelvin": 2}), "error"),
+        (({"delta_degC": 1}, {"degF": 1}), "error"),
+        (({"delta_degC": 1}, {"degC": 1}), "error"),
     ]
 
     @pytest.mark.parametrize(("input_tuple", "expected"), convert_with_offset)
@@ -985,7 +1035,7 @@ class TestConvertWithOffset(QuantityTestCase):
             assert ureg.Unit(a) == ureg.Unit("canonical")
 
         # Test that aliases defined multiple times are not duplicated
-        assert ureg._units["canonical"].aliases == (
+        assert internal(ureg)._units["canonical"].aliases == (
             "alias1",
             "alias2",
         )
